window.SynchwebShippingHandler={
		
		/**
		 * The base URI of the SynchWeb instance. 
		 * Populated from the shipmentdestination's baseuri property.
		 */
		baseUri:null,
		
		/**
		 * The authentication token to be passed to the synchrotron with each request after authentication.
		 * Received from the synchrotron on authentication.
		 */
		token:null,

		safetyLevels:[
			{ "code":"Green", "risk":"Low", "background":"#cfc", "color":"#060"},
			{ "code":"Yellow", "risk":"Medium", "background":"#ffc", "color":"#660"},
			{ "code":"Red", "risk":"High", "background":"#fcc", "color":"#600"}
		],

		
		/**
		 * Begins the process of authenticating, choosing a proposal/session, and submitting to 
		 * the synchrotron. 
		 * 
		 * It is assumed that local validation of the shipment has already been done! This includes 
		 * basic checks such as: at least one dewar; each dewar has at least one puck; all crystals
		 * have protein acronyms. Some validation can only be done at the synchrotron, but we should
		 * not be sending them shipments that fail these basic checks.
		 */
		begin:function(){
			if(!window.shipmentDestination){ 
				alert("No shipment destination set - cannot determine shipping API location.");
				return false;
			}
			var baseUri=window.shipmentDestination.baseuri;
			if(!baseUri){ 
				alert("No shipping API base URI set for "+window.shipmentDestination.name+".");
				return false;
			}
			SynchwebShippingHandler.baseUri=baseUri.replace(/\/+$/g,"")+"/";
			ui.modalBox({ title:"Send shipment to "+window.shipmentDestination.name, content:"Getting list of proposals from "+window.shipmentDestination.name+"..." });
			SynchwebShippingHandler.getProposals();
		},

		/**
		 * Renders the login form into the modal box.
		 */
		showLoginForm:function(){
			$("modalBox").down(".boxbody").innerHTML="";
			ui.setModalBoxTitle("Authenticate at "+window.shipmentDestination.name);
			var f=$("modalBox").form({
				action:SynchwebShippingHandler.baseUri+"authenticate",
				method:"post",
			})
			f.style.maxWidth="800px";
			var h=f.formField({
				label:"Authenticate with your "+window.shipmentDestination.name+" credentials", content:'&nbsp;'
			});
			h.addClassName("radiohead");
			f.textField({
				name:'remoteusername',
				label:window.shipmentDestination.name+" username",
				value:""
			});
			f.passwordField({
				name:'remotepassword',
				label:window.shipmentDestination.name+" password",
				value:""
			});
			fieldValidations.remoteusername="required";
			fieldValidations.remotepassword="required";
	
			f.submitButton({ label:"Authenticate" })
			f.onsubmit=function(){ SynchwebShippingHandler.authenticate(); return false; }
			$("remoteusername").focus();
		},
		
		/**
		 * Validates that both username and password are present, then submits them to the remote API.
		 * Expects a token in return.
		 */
		authenticate:function(){
			var frm=$("modalBox").down(".boxbody form");
			var isValid=true;
			frm.select("input").each(function(f){
				if(!validator.validate(f)){ isValid=false; }
			});
			if(!isValid){ return false; }
			var username=frm.remoteusername.value;
			var password=frm.remotepassword.value;
			frm.down("input[type=submit]").up("label").addClassName("updating");
			AjaxUtils.remoteAjax(
					SynchwebShippingHandler.baseUri+"authenticate",
					'post',
					'{ "login":"'+username+'", "password":"'+password+'" }', //Note SynchWeb wants JSON as a string, not an object 
					SynchwebShippingHandler.authenticate_onSuccess,
					SynchwebShippingHandler.authenticate_onFailure
			);
			return false;
		},
		/**
		 * Success handler for remote synchrotron authentication. Calls getProposals().
		 * Note that Synchweb returns a 200 OK with the "correct" response code in a status value (i.e., 
		 * transport.responseJSON.status instead of transport.status, which is always 200). We check for
		 * an "error" value and handle anything with it present as a failure.
		 * @param XMLHttpRequest transport The response object.
		 */
		authenticate_onSuccess:function(transport){
			if(!transport.responseJSON || !transport.responseJSON.jwt){
				if(transport.responseJSON.error){
					SynchwebShippingHandler.authenticate_onFailure(transport);
				} else {
					SynchwebShippingHandler.showErrorAtAuthenticate("Bad response from remote server. Cannot authenticate");
				}
			} else {
				SynchwebShippingHandler.token=transport.responseJSON.jwt;
				window.setTimeout(SynchwebShippingHandler.getProposals,50);
			}
		},
		/**
		 * Failure handler for remote synchrotron authentication.
		 * @param XMLHttpRequest transport The response object.
		 */
		authenticate_onFailure:function(transport){
			var msg="Could not log you in. The remote server gave no further information.";
			if(transport.responseJSON && transport.responseJSON.error){
				msg=window.shipmentDestination.name+" said: "+transport.responseJSON.error;
			}
			SynchwebShippingHandler.showErrorAtAuthenticate(msg);
		},
		/**
		 * Renders the authentication error message below the form.
		 * @param String msg The error message.
		 */
		showErrorAtAuthenticate:function(msg){
			var frm=$("modalBox").down(".boxbody").down("form");
			frm.down("input[type=submit]").up("label").removeClassName("updating");
			if(!$("autherror")){
				frm.innerHTML+='<div style="text-align:left" id="autherror" class="errorbar"></div>';
			}
			$("autherror").innerHTML=msg;
		},


		/**
		 * Fetches a list of the user's proposals from the synchrotron.
		 * On failure, triggers authentication.
		 */
		getProposals:function(){
			if(!SynchwebShippingHandler.baseUri){ 
				alert("No base URI supplied. Call begin() first, with base URI."); 
				return false;
			}
			if(!SynchwebShippingHandler.token){ 
				SynchwebShippingHandler.showLoginForm();
				return false;
			}
			AjaxUtils.remoteAjax(
					SynchwebShippingHandler.baseUri+'proposal?page=1&per_page=50&all=1&next=1',
					'get',
					{},
					SynchwebShippingHandler.getProposals_onSuccess,
					SynchwebShippingHandler.getProposals_onFailure,
					{ 'Authorization':'Bearer '+SynchwebShippingHandler.token }
			);
		},
		getProposals_onSuccess:function(transport){
			if(transport.responseJSON.error){
				//Synchweb returns 200 OK despite error
				return SynchwebShippingHandler.getProposals_onFailure(transport);
			}
			SynchwebShippingHandler.renderProposalList(transport.responseJSON);
		},
		getProposals_onFailure:function(transport){
			var msg="Error getting list of proposals";
			if(transport.responseJSON && transport.responseJSON.error){
				msg=window.shipmentDestination.name+" said: "+transport.responseJSON.error;
			}
			var mb=($("modalBox").down(".boxbody"));
			ui.setModalBoxTitle("Could not retrieve list of proposals from "+window.shipmentDestination.name);
			mb.innerHTML=msg;
		},
		renderProposalList:function(response){
			var mb=$("modalBox").down(".boxbody");
			if(response.total==0){
				ui.setModalBoxTitle("No proposals at "+window.shipmentDestination.name);
				mb.innerHTML='No proposals found. Cannot submit shipment.';
				return false;
			}
			ui.setModalBoxTitle("Your proposals at "+window.shipmentDestination.name);
			mb.innerHTML='';
			response.data.each(function(p){
				var prop=mb.treeItem({
					record:p,
					header:p.PROPOSAL+": "+p.TITLE,
					url:SynchwebShippingHandler.baseUri+'proposal/visits?page=1&per_page=50&prop='+p.PROPOSAL,
					headers:["Visit","Beamline","Active","Start","End",""],
					cellTemplates:["{{VISIT}}", "{{BL}}", [ui.checkmark,"ACTIVE", ], "{{ST}}", "{{EN}}", [SynchwebShippingHandler.getSendButton,"ACTIVE"] ],
					requestHeaders:{ 'Authorization':'Bearer '+SynchwebShippingHandler.token }
				});
				SynchwebShippingHandler.getProposalLabContacts(prop);
			});
		},
		
		getProposalLabContacts:function(elem){
			var p=elem.record;
			if(!p){ return; }
			AjaxUtils.remoteAjax(
					SynchwebShippingHandler.baseUri+"contact?page=1&per_page=100&all=1&prop="+p.PROPOSAL,
					'get',
					'', 
					function(transport){ SynchwebShippingHandler.getProposalLabContacts_onSuccess(transport,elem); },
					function(transport){ SynchwebShippingHandler.getProposalLabContacts_onFailure(transport,elem); },
					{ 'Authorization':'Bearer '+SynchwebShippingHandler.token }
			);
		},
		getProposalLabContacts_onSuccess:function(transport, elem){
			if(!transport.responseJSON || !transport.responseJSON.data){
				return SynchwebShippingHandler.getProposalLabContacts_onFailure(transport,elem);
			}
			elem.labContacts=transport.responseJSON.data;
		},
		getProposalLabContacts_onFailure:function(transport, elem){
			elem.labContacts=[];
		},

		/**
		 * Returns the HTML for a "send shipment" button.
		 * @param Object record A visit object from Synchweb.
		 */
		getSendButton:function(record){
			if(1==record.ACTIVE){
				return '<input type="button" value="Use this visit" onclick="SynchwebShippingHandler.chooseLabContact(this)" />';
			}
			//Inactive visits - currently allow and let syncthrotron barf if unhappy.
			return '<input type="button" value="Use this visit" onclick="SynchwebShippingHandler.chooseLabContact(this)" />';
		},
		
		/**
		 * Triggers final submission of the shipment to the synchrotron (along with any preflight validation).
		 * @param DOMElement btn The "Send shipment" button.
		 */
		chooseLabContact:function(btn){
			var visit=btn.up("tr").rowData;
			var proposalElement=btn.up("div.treeitem");
			var proposal=proposalElement.record;
			var labContacts=proposalElement.labContacts;
			if(undefined==labContacts){
				alert("Still waiting for list of lab contacts for this proposal. Try again in a moment.");
				return false;
			} else if(0==labContacts.length){
				alert("This proposal has no lab contacts. You need to define at least one lab contact in "+window.shipmentDestination.name+"'s ISPyB.");
				return false;
			} else {
				var box=$("modalBox").down(".boxbody");
				box.proposal=proposal;
				box.visit=visit;
				box.innerHTML='';
				box.table({
					headers:[ 'Name','Organisation',''],
					cellTemplates:[ '{{CARDNAME}}','{{LABNAME}}','<input type="button" value="Choose" onclick="SynchwebShippingHandler.setLabContact(this)" />' ],
					contentBefore:'Choose the lab contact for the shipment:'
				}, {'total':labContacts.length, 'rows':labContacts });
				ui.setModalBoxTitle("Choose the lab contact for this shipment");
			}
		},
		setLabContact:function(btn){
			var box=$("modalBox").down(".boxbody");
			box.labContact=btn.up("tr").rowData;
			SynchwebShippingHandler.chooseSafetyLevel();
		},

		chooseSafetyLevel:function(){
			ui.setModalBoxTitle("Choose the safety level for this shipment");
			var box=$("modalBox").down(".boxbody");
			box.innerHTML='';
			var t=box.table({
				headers:[ 'Colour code','Risk level',''],
				cellTemplates:[ '{{code}}','{{risk}}','<input type="button" value="Choose" onclick="SynchwebShippingHandler.setSafetyLevel(this)" />' ],
				contentBefore:'Choose the safety level for the shipment:'
			}, {'total':3, 'rows':SynchwebShippingHandler.safetyLevels });
			t.select("tr").each(function(tr){
				if(!tr.rowData){ return; }
				tr.style.backgroundColor=tr.rowData.background;
				tr.down("td").style.color=tr.rowData.color;
				tr.down("td").style.fontWeight="bold";
			});
		},
		setSafetyLevel:function(btn){
			var box=$("modalBox").down(".boxbody");
			var tr=btn.up("tr");
			box.safetyLevel=tr.rowData["code"];
			SynchwebShippingHandler.showFinalConfirmation();
		},
		
		showFinalConfirmation:function(){
			var box=$("modalBox").down(".boxbody");
			var proposal=box.proposal;
			var visit=box.visit;
			var labContact=box.labContact;
			box.innerHTML='';
			ui.setModalBoxTitle("Check the details before sending the shipment");
			var f=box.form({ action:'#', method:'post' });
			f.setStyle({ position:"absolute", top:"1%", left:"1%", width:"48%" });
			f.textField({ readonly:true, label:'Destination', value:window.shipmentDestination.name });
			f.textField({ readonly:true, label:'Proposal', value:proposal.PROPOSAL });
			f.textField({ readonly:true, label:'Visit', value:visit.VISIT });
			f.textField({ readonly:true, label:'Home lab contact', value:labContact.CARDNAME });
			
			SynchwebShippingHandler.safetyLevels.each(function(sl){
				if(sl.code!=box.safetyLevel){ return; }
				var ff=f.textField({ readonly:true, label:'Safety level', value:sl.code });
				ff.style.backgroundColor=sl.background;
			});

			var f2=box.form({ action:'#', method:'post' });
			f2.setStyle({ position:"absolute", top:"1%", right:"1%", width:"48%", marginTop:0, paddingTop:0 });
			var b=f2.formField({ readonly:true, label:'Review this information to make sure that it is correct, then click "Send shipment".', content:'<input type="button" value="Send shipment" onclick="SynchwebShippingHandler.sendShipment()" />' });
			b.setStyle({ "height":f.getHeight()+"px"});
			b.down("input").setStyle({ position:"absolute", bottom:"0.5em", right:"0.5em"});

			box.innerHTML+='<div style="position:absolute;bottom:10%; left:1%; width:98%; text-align:center"><a href="#" onclick="ui.closeModalBox();return false">Cancel - don\'t send this shipment</a></div>';
		},
		
		sendShipment:function(){
			var box=$("modalBox").down(".boxbody");
			var proposal=box.proposal;
			var visit=box.visit;
			var labContact=box.labContact;
			if(!confirm("Really submit shipment to "+window.shipmentDestination.name+", visit "+visit.VISIT+"?")){ return false; }
			ui.setModalBoxTitle("Submitting your shipment to "+window.shipmentDestination.name+"...");
			var box=$("modalBox").down(".boxbody");
			box.innerHTML='Beginning to send shipment...';
			SynchwebShippingHandler.getProteins();
		},
		
		getProteins:function(){
			var box=$("modalBox").down(".boxbody");
			var proposal=box.proposal;
			AjaxUtils.remoteAjax(
					SynchwebShippingHandler.baseUri+"sample/proteins?page=1&per_page=1000&all=1&prop="+proposal.PROPOSAL,
					'get',
					'', 
					SynchwebShippingHandler.doPreflight,
					SynchwebShippingHandler.getProteins_onFailure,
					{ 'Authorization':'Bearer '+SynchwebShippingHandler.token }
			);			
		},
		getProteins_onFailure:function(transport){
			ui.setModalBoxTitle("Could not send shipment");
			var box=$("modalBox").down(".boxbody");
			if(!transport.responseJSON){
				box.innerHTML='Could not get list of proteins from '+window.shipmentDestination.name+', so could not validate shipment before sending.<br/>Please try again later.';
			} else if(transport.responseJSON.message){
				box.innerHTML='Could not get list of proteins from '+window.shipmentDestination.name+', so could not validate shipment before sending.<br/>The remote server said: '+transport.responseJSON.message+'<br/>Please try again later.';
			} else if(0==transport.responseJSON.total){
				box.innerHTML='Proposal  '+box.proposal.PROPOSAL+', has no proteins. Therefore none of your samples have a protein acronym that appears in the proposal, and the shipment cannot be sent.<br/>Create the proteins in the '+window.shipmentDestination.name+' ISPyB system.';
			}
		},
		
		doPreflight:function(transport){
			if(!transport.responseJSON || 0==transport.responseJSON.total || 400<=transport.responseJSON.status){
				return SynchwebShippingHandler.getProteins_onFailure(transport);
			}
			var box=$("modalBox").down(".boxbody");
			var acronyms=[];
			transport.responseJSON.data.each(function(prot){
				acronyms.push(prot.ACRONYM);
			});
			var errors=Shipment.getShipmentErrors(acronyms);
			
			if(0<errors.length){
				box.innerHTML="Could not send shipment. Please fix the following:<br/><ul><li>"+errors.join("</li><li>")+"</ul>";
				return false;
			}
			var dewars=$$(".containertab");

			dewars.each(function(tab){
				var dewarName=tab.dataset.containername;
				var nameFormat=new RegExp('^DLS-MX-\d\d\d\d$')
				if(!nameFormat.test(dewarName)){
					errors.push('Dewar name does not match Diamond barcode format DLS-MX-nnnn');
				}
				tab.addClassName("noRemoteDewar");
			});

			box.innerHTML+="<br/>Pre-flight validation complete. Creating shipment in ISPyB...";
			SynchwebShippingHandler.getOrCreateDewars();
			SynchwebShippingHandler.remoteErrors=[];
		},

		remoteErrors:[],
		logRemoteError:function(err){
			SynchwebShippingHandler.remoteErrors.push(msg);
		},
		showRemoteErrors:function(){
			alert("showRemoteErrors")
			var box=$("modalBox").down(".boxbody");
			box.innerHTML+='The following errors occurred:<br/><br/>';
			box.innerHTML+=SynchwebShippingHandler.remoteErrors.join('<br/>');
		},

		//Call this until all tab.hasRemoteDewar are true
		getOrCreateDewars:function(){
			var box=$("modalBox").down(".boxbody");
			var labContact=box.labContact;
			var proposal=box.proposal;
			var visit=box.visit;
			var dewars=$$(".noRemoteDewar");
			if(SynchwebShippingHandler.remoteErrors.length>0){
				//Something went wrong on a previous call
				SynchwebShippingHandler.showRemoteErrors();
				return;
			} else if(!dewars || 0==dewars.length){
				//Move on to the next stage
				box.innerHTML+="<br/>All dewars exist at synchrotron. Continuing.";
				alert(SynchwebShippingHandler.submitBasicShipmentDetails);
				SynchwebShippingHandler.submitBasicShipmentDetails();
				alert("after");
				return;
			}
			var tab=dewars[0];
			var dewarName=tab.dataset.containername;
			AjaxUtils.remoteAjax(
				SynchwebShippingHandler.baseUri+"shipment/dewars/registry/"+dewarName+"?prop="+proposal.PROPOSAL,
				'get',
				'', 
				function(transport){ SynchwebShippingHandler.getOrCreateDewars_onFailure(transport, tab)},
				function(transport){ SynchwebShippingHandler.getOrCreateDewars_onFailure(transport, tab)},
				{ 'Authorization':'Bearer '+SynchwebShippingHandler.token }
			);
		},
		getOrCreateDewars_onSuccess:function(transport, tab){
			if(transport.responseJSON.status && 400<=transport.responseJSON.status){
				return SynchwebShippingHandler.getOrCreateDewars_onFailure(transport,tab);
			}
			//Found an existing dewar
			var box=$("modalBox").down(".boxbody");
			var dewarName=tab.dataset.containername;
			box.innerHTML+="<br/>Dewar "+dewarName+" already exists in ISPyB";
			tab.remoteDewar=transport.responseJSON;
			tab.removeClassName("noRemoteDewar");
			//and go round again
			SynchwebShippingHandler.getOrCreateDewars();
		},
		getOrCreateDewars_onFailure:function(transport,tab){
			//Dewar was not found, for some reason
			var box=$("modalBox").down(".boxbody");
			var labContact=box.labContact;
			var proposal=box.proposal;
			var dewarName=tab.dataset.containername;
//			if("400"==transport.responseJSON.status){
				//"400" with a 200 OK means "not found" in Synchweb-land... I think...
				box.innerHTML+="<br/>Dewar "+dewarName+" not found in ISPyB, creating it";
				AjaxUtils.remoteAjax(
						SynchwebShippingHandler.baseUri+"shipment/dewars/registry/"+dewarName,
						'put',
						'{"new":true,"FACILITYCODE":"'+dewarName+'","LABCONTACTID":"'+labContact.LABCONTACTID+'","PURCHASEDATE":"","prop":"'+proposal.PROPOSAL+'"}',
						function(transport){ SynchwebShippingHandler.createDewar_onSuccess(transport,tab); },
						function(transport){ SynchwebShippingHandler.createDewar_onFailure(transport,tab); },
						{ 'Authorization':'Bearer '+SynchwebShippingHandler.token }
				);
//			} else {
//				//Something went wrong.
//				SynchwebShippingHandler.logRemoteError('Dewar '+dewarName+' was not found at '+window.shipmentDestination.name+' and could not be created.');
//			}
		},
		createDewar_onSuccess:function(transport,tab){
			if(transport.responseJSON.status && 400<=transport.responseJSON.status){
				return SynchwebShippingHandler.createDewar_onFailure(transport,tab);
			}
			var box=$("modalBox").down(".boxbody");
			//Dewar was not found but we created it
			tab.remoteDewar=transport.responseJSON;
			tab.removeClassName("noRemoteDewar");
			box.innerHTML+="<br/>Dewar "+dewarName+" created in ISPyB";
			//and go round again because created response is not the same as getting existing
			SynchwebShippingHandler.getOrCreateDewars();
		},
		createDewar_onFailure:function(transport,tab){
			//Dewar was not found and we could not create it
			var dewarName=tab.dataset.containername;
			SynchwebShippingHandler.logRemoteError('Dewar '+dewarName+' was not found at '+window.shipmentDestination.name+' and could not be created.');
			//and go round again to trigger error dumping
			SynchwebShippingHandler.getOrCreateDewars();
		},
		
		submitBasicShipmentDetails:function(){
			var box=$("modalBox").down(".boxbody");
			box.innerHTML+="<br/>Submitting basic shipment details";
			var safetyLevel=box.safetyLevel;
			var proposal=box.proposal.PROPOSAL;
			var visitId=box.visit.SESSIONID;
			var labContactId=box.labContact.LABCONTACTID;
			var shipmentName=$("shipmentdetails_form").down("#name").value;
			
			//This appears to be pointless, since the shipment gets created with no dewars anyway as far as I can tell.
			var dewars=$$("containertab");
			var dewarBarcodes=[];
			dewars.each(function(tab){
				dewarBarcodes.push(tab.dataset.containername);
			});
			numBarcodes=dewarBarcodes.length;
			dewarBarcodes='["'+dewarBarcodes.join('","')+'"]';
			
			AjaxUtils.remoteAjax(
					SynchwebShippingHandler.baseUri+"shipment/shipments",
					'post',
					'{"SHIPPINGNAME":"'+shipmentName+'","DEWARS":"'+numBarcodes+'","FCODES":'+dewarBarcodes+',"FIRSTEXPERIMENTID":"'+visitId+'","noexp":null, '+
						'"SAFETYLEVEL":"'+safetyLevel+'","COMMENT":"","SENDINGLABCONTACTID":"'+labContactId+'","RETURNLABCONTACTID":"'+labContactId+'", '+
						'"DELIVERYAGENT_SHIPPINGDATE":"","PHYSICALLOCATION":"","READYBYTIME":"","CLOSETIME":"","DELIVERYAGENT_DELIVERYDATE":"", '+
						'"DELIVERYAGENT_AGENTNAME":"","DELIVERYAGENT_AGENTCODE":"","prop":"'+proposal+'" '+
					'}',
					SynchwebShippingHandler.submitBasicShipmentDetails_onSuccess,
					SynchwebShippingHandler.submitBasicShipmentDetails_onFailure,
					{ 'Authorization':'Bearer '+SynchwebShippingHandler.token }
			);
		},
		submitBasicShipmentDetails_onSuccess:function(transport){
			if(transport.responseJSON && transport.responseJSON.SHIPMENTID){
				box.shipmentId=transport.responseJSON.SHIPMENTID;
			}
			SynchwebShippingHandler.addDewarsToShipment();
		},
		submitBasicShipmentDetails_onFailure:function(transport){
			SynchwebShippingHandler.logRemoteError("Could not create shipment.");
			SynchwebShippingHandler.showRemoteErrors();
		},
		
		/**
		 * After creating the shipment, and with all dewars existing at the synchrotron, add them to the shipment
		 */
		addDewarsToShipment:function(){
			var box=$("modalBox").down(".boxbody");
			box.innerHTML+="<br/>Adding dewars to shipment";
			var dewars=$$("containertab");
			var dewarBarcodes=[];
			dewars.each(function(tab){
				tab.addedToShipment=false;
				dewarBarcodes.push(tab.dataset.containername);
			});
			
		},
		addDewarsToShipment_onSuccess:function(transport){
			
			var dewars=$$("containertab");
				
		
		},
		addDewarsToShipment_onFailure:function(transport){
		},
		
}